import matplotlib.pyplot as plt 
import seaborn as sns 
import pandas as pd

class Plot:
    def __init__(self, data, metric):
        self.data = data
        self.metric_name = metric
        self.plot_data = {}
        self.non_plot_data = {}
        self.disppearing_ratio = 500 # The ratio at which to remove outliers (too large/small values)

    def calculate(self):

        for model in self.data: # model -> name of model

            if self.metric_name in self.data[model] and self.data[model][self.metric_name] == None: # If predict_proba = True and model has no predict_proba property (no y_pred)
                self.non_plot_data[model] = None
            else:
                self.plot_data[model] = self.data[model][self.metric_name]

        return self.plot_data

    def display(self):
        self.plot_data = dict(sorted(self.plot_data.items(), key=lambda item: item[1]* -1))

        # We have to filter out those values where it is too big/small so that they don't skew the plot O(N)
        keys = list(self.plot_data.keys())
        for i, model in enumerate(self.plot_data):
            val = self.plot_data[model]
            if i < len(self.plot_data)-1:
                next_val = self.plot_data[keys[i+1]]
            else:
                break

            if abs(val/next_val) > self.disppearing_ratio:
                self.non_plot_data[model] = val

        for model in self.non_plot_data:
            del self.plot_data[model]

        # Making a dataframe to use with plotly
        df = pd.DataFrame({'Model':list(self.plot_data.keys()), 'Values':list(self.plot_data.values())})
    
        sns.set_theme()
        sns.set_context("paper")
        ax = sns.barplot(df, x='Values', y='Model').set(title='Title of Plot')
        ax.bar_label(ax.containers[0])

        # Return non plottable data
        if len(self.non_plot_data) > 0:
            print(f"There were {len(self.non_plot_data)} number of non plottable data")

        plt.show()
        return self.non_plot_data if len(self.non_plot_data) > 0 else "No data points were removed"



"""
# Testing code

hi = {'Linear Regression': 5.945477578686267e+30, 'Gaussian Process Regressor': 38862842948.98595, 'MLP Regressor': 35413863580.22882, 'SVR': 7197385705.765231, 'Perceptron': 3634079418.0570774, 'Decision Tree Regressor': 2015704715.13242, 'ElasticNet': 1139044333.7945106, 'Ada Boost Regressor': 1114901189.166725, 'SGDRegressor': 963807765.7846137, 'ARDRegression': 881823520.212382, 'Lasso': 863589220.5745481, 'Bayesian Ridge': 862270354.5060233, 'Ridge': 814237316.0583718, 
'XGBRegressor': 757239500.2347634, 'Random Forest Regressor': 739429472.215051, 'Passive Aggressive Regressor': 731582371.3055538, 'Gradient Boosting Regressor': 633820421.5121104}
tmp = Plot(hi, 'mean_squared_error')
tmp.display()

"""



from sklearn.metrics import accuracy_score, balanced_accuracy_score, average_precision_score, brier_score_loss, f1_score, log_loss, precision_score, recall_score, jaccard_score, roc_auc_score, explained_variance_score, max_error, mean_absolute_error, mean_squared_error, mean_squared_log_error, median_absolute_error, r2_score, mean_absolute_percentage_error, d2_absolute_error_score, d2_pinball_score, d2_tweedie_score
from sklearn.preprocessing import minmax_scale
from tqdm import tqdm

class Score():
    def __init__(self, data, OutputType):
        self.res = {} # dict{Name-of-model: {metrics_name:metrics_val} }
        self.data = data
        self.task = OutputType
        # data is in the type {name-of-model : [y_true, y_pred]}

    def calculate(self):
        keys = list(self.data.keys())
        if self.task == 'regression':
            for i in tqdm(range(len(self.data))): # Looping over different models 
                key = keys[i] # key => model names 
                y_true, y_pred = self.data[key][0], self.data[key][1]
                self.res[key] = {}
                if y_pred != None:
                    self.res[key]['explained_variance'] = explained_variance_score(y_true, y_pred)
                    self.res[key]['max_error'] = max_error(y_true, y_pred)
                    self.res[key]['mean_absolute_error'] = mean_absolute_error(y_true, y_pred)
                    self.res[key]['mean_squared_error'] = mean_squared_error(y_true, y_pred)
                    self.res[key]['mean_squared_log_error'] = mean_squared_log_error(minmax_scale(y_true, feature_range=(0,1)), minmax_scale(y_pred, feature_range=(0,1)))
                    self.res[key]['median_absolute_error'] = median_absolute_error(y_true, y_pred)
                    self.res[key]['r2_score'] = r2_score(y_true, y_pred)
                    self.res[key]['mean_absolute_percentage_error'] = mean_absolute_percentage_error(y_true, y_pred)
                    self.res[key]['d2_absolute_error_score'] = d2_absolute_error_score(y_true, y_pred)
                    self.res[key]['d2_pinball_score'] = d2_pinball_score(y_true, y_pred)
                    self.res[key]['d2_tweedie_score'] = d2_tweedie_score(y_true, y_pred)

        else: # Task is classification 
            for i in tqdm(range(len(self.data))): # Looping over different models 
                key = keys[i]
                y_true, y_pred = self.data[key][0], self.data[key][1]
                self.res[key] = {}

                # Calculating possible scores
                # try;except for models where predict_proba doesn't work -> in this case the score would reflect None
                # For None scores, we won't plot them 
                try: # In case y_pred is None
                    
                    try: # These do not support decimal point prediction values (no support for predict_proba)
                        self.res[key]['accuracy_score'] = accuracy_score(y_true, y_pred)
                    except:
                        self.res[key]['accuracy_score'] = None
                    
                    try:
                        self.res[key]['balanced_accuracy_score'] = balanced_accuracy_score(y_true, y_pred)
                    except:
                        self.res[key]['balanced_accuracy_score'] = None                    

                    try:
                        self.res[key]['f1_score'] = f1_score(y_true, y_pred)
                    except:
                        self.res[key]['f1_score'] = None

                    try:
                        self.res[key]['precision_score'] = precision_score(y_true, y_pred)
                    except:
                        self.res[key]['precision_score'] = None

                    try:
                        self.res[key]['recall_score'] = recall_score(y_true, y_pred)
                    except:
                        self.res[key]['recall_score'] = None

                    try:
                        self.res[key]['jaccard_score'] = jaccard_score(y_true, y_pred)
                    except:
                        self.res[key]['jaccard_score'] = None

                    self.res[key]['roc_auc_score'] = roc_auc_score(y_true, y_pred)
                    self.res[key]['log_loss'] = log_loss(y_true, y_pred)
                    self.res[key]['average_precision_score'] = average_precision_score(y_true, y_pred)
                    self.res[key]['brier_score_loss'] = brier_score_loss(y_true, y_true)

                except:
                    continue

        return self.res



import numpy as np 
import pandas as pd
from sklearn import preprocessing
from sklearn.preprocessing import StandardScaler, OneHotEncoder
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import make_pipeline, Pipeline
from sklearn.model_selection import train_test_split
from sklearn.impute import SimpleImputer
import xgboost as xgb
from sklearn.linear_model import PassiveAggressiveClassifier, LogisticRegression, SGDClassifier, RidgeClassifier, LinearRegression, Ridge, Lasso, ElasticNet, BayesianRidge, ARDRegression, SGDRegressor, PassiveAggressiveRegressor, Perceptron
from sklearn.gaussian_process import GaussianProcessRegressor, GaussianProcessClassifier
from sklearn import svm
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.ensemble import GradientBoostingClassifier, AdaBoostClassifier, RandomForestClassifier, RandomForestRegressor, AdaBoostRegressor, GradientBoostingRegressor
from sklearn.neural_network import MLPRegressor, MLPClassifier
from tqdm import tqdm

from easy_ml.metrics.metrics import *
from easy_ml.plot.plot import *


class ExploreModel:
    def __init__(self, data, FeatureList, Target, PredictProba = False, EnsembleBoolean=True, NeuralBoolean=True, SplitRatio=0.3, OutputType='regression', Randomstate=42):
        # Any possible parameters 
        self.PredictProba = PredictProba
        if not type(self.PredictProba) is bool:
            raise TypeError("Only Boolean variables accepted for PredictProba parameter")
        self.OutputType = OutputType
        if self.OutputType not in ['regression', 'classification']:
            raise Exception("OutputType should either be 'regression' or 'classification'")

        self.FeatureList = FeatureList
        if len(self.FeatureList) <= 0:
            raise Exception("There should be 1 or more features")

        self.Target = Target
        if self.Target not in data.columns.tolist():
            raise Exception("The target variable is not in the DataFrame")

        self.EnsembleBoolean = EnsembleBoolean
        if not type(self.EnsembleBoolean) is bool:
            raise TypeError("Only Boolean variables accepted for Ensemble Boolean parameter")

        self.NeuralBoolean = NeuralBoolean   
        if not type(self.NeuralBoolean) is bool:
            raise TypeError("Only Boolean variables accepted for Neural Boolean parameter")

        self.df = data # In a pandas DataFrame
        self.SplitRatio = SplitRatio
        if self.SplitRatio > 1 or self.SplitRatio < 0:
            raise Exception("Split Ratio should be between 0 and 1")

        self.RandomState = Randomstate
        self.PredictedVal = {}
        self.Scores = {}
        self.outlier_values = {}

        self.Regression_Models = [['Linear Regression', LinearRegression()], ['Ridge', Ridge()], ['Lasso', Lasso()], ['ElasticNet', ElasticNet()], 
        ['Bayesian Ridge', BayesianRidge()], ['ARDRegression', ARDRegression()], ['SGDRegressor', SGDRegressor()], 
        ['Passive Aggressive Regressor', PassiveAggressiveRegressor()], ['Perceptron', Perceptron()], ['SVR', svm.SVR()], ['Gaussian Process Regressor', GaussianProcessRegressor()], 
        ['Decision Tree Regressor', DecisionTreeRegressor()]]
        self.Ensemble_Regression_Models = [['Random Forest Regressor', RandomForestRegressor()], ['Ada Boost Regressor', AdaBoostRegressor()],
        ['Gradient Boosting Regressor', GradientBoostingRegressor()], ['XGBRegressor', xgb.XGBRegressor()]]
        self.Neural_Regression_Models = [['MLP Regressor', MLPRegressor()]]

        self.Classification_Models = [['Ridge Classifier', RidgeClassifier()], ['SGD Classifier', SGDClassifier()], ['Logistic Regression', LogisticRegression()], 
        ['Passive Agressive Classifier', PassiveAggressiveClassifier()], ['SVC', svm.SVC()], ['KNN Classifier', KNeighborsClassifier()], ['Gaussian Process Classifier', GaussianProcessClassifier()],
        ['GaussianNB', GaussianNB()], ['Decision Tree Classifier', DecisionTreeClassifier()]]
        self.Ensemble_Classification_Models = [['Random Forest Classifier', RandomForestClassifier()], ['Ada Boost Classifier',AdaBoostClassifier()],
        ['Gradient Boosting Classifier', GradientBoostingClassifier()]]
        self.Neural_Classification_Models = [['MLP CLassifier', MLPClassifier()]]

    def __prepare_data(self): # Private Method 
        df = self.df[self.FeatureList].copy()

        numerical_df = df.select_dtypes(include=['int', 'float'])
        categorical_df = df.select_dtypes(exclude = ['int', 'float']) # Contains object, boolean

        num_pipe = make_pipeline(
            SimpleImputer(strategy='median'),
            StandardScaler()
        )
        # pipeline for categorical columns
        cat_pipe = make_pipeline(
            SimpleImputer(strategy='constant', fill_value='N/A'),
            OneHotEncoder(handle_unknown='ignore', sparse_output=False)
        )

        # combine both the pipelines
        full_pipe = ColumnTransformer([
            ('num', num_pipe, numerical_df.columns.tolist()),
            ('cat', cat_pipe, categorical_df.columns.tolist())
        ])

        return full_pipe

    def __split(self):
        X = self.df[self.FeatureList].copy()
        y = self.df[self.Target].copy()

        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=self.SplitRatio, random_state=self.RandomState)

        return X_train, X_test, y_train, y_test

    def fit(self):
        full_pipe = self.__prepare_data()

        X_train, X_test, y_train, y_test = self.__split()

        if self.OutputType == 'regression':
            if self.EnsembleBoolean:
                if self.NeuralBoolean:
                    models = self.Regression_Models + self.Ensemble_Regression_Models + self.Neural_Regression_Models 
                else:
                    models = self.Regression_Models + self.Ensemble_Regression_Models
            else:
                if self.NeuralBoolean:
                    models = self.Regression_Models + self.Neural_Regression_Models
                else:
                    models = self.Regression_Models
        else: # self.OutputType == 'classification'
            if self.EnsembleBoolean:
                if self.NeuralBoolean:
                    models = self.Classification_Models + self.Ensemble_Classification_Models + self.Neural_Classification_Models
                else:
                    models = self.Classification_Models + self.Ensemble_Classification_Models
            else:
                if self.NeuralBoolean:
                    models = self.Classification_Models + self.Neural_Classification_Models
                else:
                    models = self.Classification_Models

        for i in tqdm(range(len(models))):
            name, model = models[i]

            # Utilize the pipeline from earlier
            model_pipeline = make_pipeline(full_pipe, model)

            # Train the model
            model_pipeline.fit(X_train, y_train)

            # Make predictions on the test set
            if self.PredictProba:
                try:
                    y_pred = model_pipeline.predict_proba(X_test)[:,1]
                except:
                    y_pred = None
            else:
                y_pred = model_pipeline.predict(X_test)

            # Add to self.PredictedVal
            self.PredictedVal[name] = [y_test, y_pred]

        return self.PredictedVal # A dictionary of name, predicted values, actual values 

    def calculate_accuracy(self):

        score = Score(self.PredictedVal, self.OutputType)    
        self.Scores = score.calculate() # Contains dict{Name-of-model: {metrics_name:metrics_val} }
        # self.Scores should be used as parameter to plot function

        return 

    def plot(self, metric): # Should plot the metrics

        plot = Plot(self.Scores, metric)
        plot.calculate()
        self.outlier_values = plot.display()

        return  

#---------------------------------------Testing---------------------------------------#

"""
df = pd.read_csv('model-exploration/classification.csv')

model = ExploreModel(df, df.columns.tolist()[2:], 'Survived', PredictProba = False, EnsembleBoolean=True, NeuralBoolean=True, SplitRatio=0.3, OutputType='classification')
model.fit()
model.calculate_accuracy()
model.plot('accuracy_score')
print(model.outlier_values)
"""





import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
    name="lazyeval", # Replace with your own username
    version="0.0.1",
    author="Kim Hyun Bin",
    author_email="KIMH0004@e.ntu.edu.sg",
    description="A python package to automate machine learning process to showcase metric values in an instance",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vanilladucky/ModelExploration",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
)